import os
import re
import sys
import errno
import inspect
import collections
from glob import glob
from datetime import datetime
from importlib import import_module

# 3rd Party
from click import echo


def generate_migration_name(folder, name):
    # Can not start with a number so starting with a underscore
    rel_path = "{folder}/_{timestamp}_{name}_migration.py".format(
        folder=folder,
        timestamp=datetime.utcnow().strftime('%Y%m%d%H%M'),
        name=name
    )
    return os.path.abspath(rel_path)

def create_package_if_necessary(dir):
    try:
        os.makedirs(dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    try:
        with open(os.path.join(os.path.abspath(dir), '__init__.py'), 'w') as f:
            f.write('# this file makes migrations a package\n')
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def find_migrations(config):
    migrations = {}
    for file in glob('{}/*_migration.py'.format(config.migration_directory)):
        migration_name = os.path.splitext(os.path.basename(file))[0]
        if sys.version_info >= (3, 3):
            from importlib import invalidate_caches
            invalidate_caches()
        migration_module = import_module("migrations.{}".format(migration_name))
        for name, obj in inspect.getmembers(migration_module):
            if inspect.isclass(obj) and re.search('Migration$', name) and name not in ['BaseMigration',
                                                                                       'MongoBackedMigration']:
                migrations[migration_name] = obj

    # 2) Ensure that the are ordered
    ordered = collections.OrderedDict(sorted(migrations.items()))
    return ordered


import os
import re
import sys
import inspect
from glob import glob
from importlib import import_module


def generate_queryset_name(folder, name):
    # Can not start with a number so starting with a underscore
    rel_path = "{folder}/{name}_queryset.py".format(
        folder=folder,
        name=name
    )
    return os.path.abspath(rel_path)


def querysets(config):
    query_sets = {}

    for file in glob('{}/*_queryset.py'.format(config.queryset_directory)):
        queryset_name = os.path.splitext(os.path.basename(file))[0]
        if sys.version_info >= (3, 3):
            from importlib import invalidate_caches
            invalidate_caches()
        queryset_module = import_module("querysets.{}".format(queryset_name))
        for name, obj in inspect.getmembers(queryset_module):
            if inspect.isclass(obj) and re.search('QuerySet$', name) and name not in ['QuerySet']:
                query_sets[name] = obj
    return query_sets


import subprocess
from tempfile import mkdtemp

import click
import pymongo
import mongoengine
from click import echo

from .utils import temp_directory
from .models import Migration, MigrationHistoryStorage
from .query_sets import querysets


def establish_datastore_connection(environment):
    mongo_db_name = environment['db_name']

    username_password_couple = ""
    if 'username' in environment:
        if 'password' in environment:
            username_password_couple = "{}:{}@".format(environment['username'], environment['password'])
        else:
            username_password_couple = "{}@".format(environment['username'])

    if 'port' in environment:
        raise Exception("port no longer supported, use the host:port syntax in the host parameter")

    host_and_port = environment['host']

    db_name = "/{}".format(environment['db_name'])

    uri = "mongodb://{username_password_couple}{host_and_port}{db_name}".format(username_password_couple=username_password_couple,
                                                                                host_and_port=host_and_port,
                                                                                db_name=db_name)

    if 'sslCAFile' in environment:
        echo('appending ssl')
        uri = "{}?ssl=true&ssl_ca_certs={}".format(uri, environment['sslCAFile'])

    echo('executing: {}'.format(uri))
    return mongoengine.connect(mongo_db_name, host=uri)


class MongoMigrationHistory(MigrationHistoryStorage, mongoengine.Document):
    """
    Mongo Table to keep track of the status of migrations
    """
    key = mongoengine.StringField()
    state = mongoengine.StringField(default=Migration.STATE_NEW)
    processed_at = mongoengine.DateTimeField()

    @classmethod
    def find_or_create_by_key(cls, migration_key):
        result = cls.objects(key=migration_key)
        if len(result) == 1:
            return result[0]
        else:
            return cls(key=migration_key).save()

    @classmethod
    def find_by_key(cls, migration_key):
        results = cls.objects(key=migration_key)
        if len(results) == 1:
            return results[0]
        else:
            return None

    @classmethod
    def all(cls):
        return cls.objects()


class MongoBackedMigration(Migration):

    def update_status(self, state):
        migration_meta = MongoMigrationHistory.find_or_create_by_key(self.migration_key)
        migration_meta.update(set__state=state)

    @property
    def status(self):
        migration_meta = MongoMigrationHistory.find_or_create_by_key(self.migration_key)
        return migration_meta.state


def dump_db(from_env, **kwargs):
    """accepts temp_dir and QuerySet as keyword options"""

    if 'temp_dir' in kwargs:
        temp_dir = kwargs['temp_dir']
    else:
        temp_dir = mkdtemp()

    if 'QuerySet' in kwargs:
        QuerySet = kwargs['QuerySet']

    echo("env: {}".format(from_env))

    options = {
        '-h': from_env['host'],
        '-d': from_env['db_name'],
        '-o': temp_dir
    }
    if 'username' in from_env:
        options['-u'] = from_env['username']
    if 'password' in from_env:
        options['-p'] = from_env['password']

    if QuerySet:
        echo("In Query Set: Env: {}".format(from_env))
        connection = establish_datastore_connection(from_env)
        database = connection[from_env['db_name']]

        query_set = QuerySet(database, options)

        query_set.execute()

    else:

        execution_array = ['mongodump']
        if 'sslCAFile' in from_env:
            execution_array.append('--ssl')
            execution_array.extend(['--sslCAFile', from_env['sslCAFile']])

        for option in options:
            execution_array.extend([option, options[option]])
        echo("Executing: {}".format(execution_array))
        subprocess.call(execution_array)

    # mongorestore -h localhost --drop -d spotlight db/backups/spotlight-staging-1/
    dump_path = "{}/{}".format(temp_dir, from_env['db_name'])
    return dump_path


def copy_db(from_env, to_env, query_set=None):
    with temp_directory() as temp_dir:
        dump_path = dump_db(from_env, temp_dir=temp_dir, QuerySet=query_set)
        restore(dump_path, to_env)


def restore(dump_path, to_env):
    drop(to_env)

    options = {
        '-h': to_env['host'],
        '-d': to_env['db_name'],
    }

    if 'username' in to_env:
        options['-u'] = to_env['username']

    if 'password' in to_env:
        options['-p'] = to_env['password']

    execution_array = ['mongorestore', '--drop']

    if 'sslCAFile' in to_env:
        execution_array.append('--ssl')
        execution_array.extend(['--sslCAFile', to_env['sslCAFile']])

    for option in options:
        execution_array.extend([option, options[option]])

    execution_array.append(dump_path)

    echo("Executing: {}".format(execution_array))
    subprocess.call(execution_array)


def drop(environ):

    options = {
        '--eval': '"db.dropDatabase()"'
    }

    database_string = "{}/{}".format(environ['host'], environ['db_name'])

    execution_array = ['mongo', database_string]

    if 'sslCAFile' in environ:
        execution_array.append('--ssl')
        execution_array.extend(['--sslCAFile', environ['sslCAFile']])

    for option in options:
        execution_array.extend([option, options[option]])

    echo()
    echo("You are about to execute the following database drop")
    echo("    {}".format(' '.join(execution_array)))
    echo()
    click.confirm('ARE YOU SURE??', abort=True)

    subprocess.call(' '.join(execution_array), shell=True)



import os
import shutil
import zipfile
from datetime import datetime

from .mongo import restore, dump_db
from .utils import temp_directory, exit_with_message, zipdir


def local_restore(zip_path, to_environment):
    zip = zipfile.ZipFile(zip_path)
    with temp_directory() as temp_dir:
        zip.extractall(path=temp_dir)
        restore(temp_dir, to_environment)


def local_backups(local_config):
    if 'backup_dir' not in local_config:
        exit_with_message('Local Settings not configured correctly, expecting "backup_dir"')

    backup_dir = local_config['backup_dir']

    if not os.path.isdir(backup_dir):
        exit_with_message('Directory [{}] does not exist.  Exiting ...'.format(backup_dir))

    backups = {}
    for item in os.listdir(backup_dir):
        backups[item] = os.path.join(backup_dir, item)

    return backups


def backup_localy(environment, local_settings, name, query_set_class=None):

    if 'backup_dir' not in local_settings:
        exit_with_message('Local Settings not configured correctly, expecting "backup_dir"')

    backup_dir = local_settings['backup_dir']

    if not os.path.isdir(backup_dir):
        exit_with_message('Directory [{}] does not exist.  Exiting ...'.format(backup_dir))

    dump_path = dump_db(environment, QuerySet=query_set_class)
    zipf = zipdir(dump_path)

    unique_file_path = generate_unique_name(backup_dir, environment, name)

    shutil.move(zipf.filename, unique_file_path)


def generate_unique_name(backup_dir, environemnt, name_prefix):
    # generate_file_name
    # database_name__2013_03_01.dmp.zip
    # or if that exists
    # database_name__2013_03_01(2).dmp.zip

    if name_prefix and name_prefix != '':
        name_base = name_prefix
    else:
        name_base = environemnt['db_name']

    name_attempt = "{}__{}.dmp.zip".format(name_base, datetime.utcnow().strftime("%Y_%m_%d"))

    # check if file exists
    name_attempt_full_path = os.path.join(backup_dir, name_attempt)

    if not os.path.exists(name_attempt_full_path):
        return name_attempt_full_path
    else:
        counter = 1
        while True:
            counter += 1
            name_attempt = "{}__{}_{}.dmp.zip".format(name_base,
                                                      datetime.utcnow().strftime("%Y_%m_%d"), counter)
            name_attempt_full_path = os.path.join(backup_dir, name_attempt)
            if os.path.exists(name_attempt_full_path):
                continue
            else:
                return name_attempt_full_path



import os
import re
import sys
import inspect
import subprocess
from copy import copy

# 3rd Party
import click
from click import echo


class Migration(object):
    """
    All migration will inherit from this.
    """
    # States
    STATE_NEW = 'New'
    STATE_PROCESSING = 'Processing'
    STATE_FAILED = 'Failed'
    STATE_COMPLETED = 'Completed'

    @property
    def migration_key(self):
        migration_file = inspect.getfile(self.__class__)
        migration_key = os.path.splitext(os.path.basename(migration_file))[0]
        return migration_key

    @property
    def migration_name(self):
        return self.__class__.__name__

    def update_status(self, state):
        raise NotImplementedError("This is an abstract class")

    @property
    def status(self):
        raise NotImplementedError("This is an abstract class")

    def process(self, force=False):
        click.echo("Processing {}".format(self.migration_name))

        if self.status == Migration.STATE_NEW or force:
            self.update_status(Migration.STATE_PROCESSING)
            echo("Starting: {}".format(self.migration_name))
            try:
                self.run()
            except Exception:
                echo("Migration {} Failed".format(self.migration_name))
                typ, value, traceback = sys.exc_info()
                echo("Unexpected error: [{}]".format(typ))
                echo("Unexpected value: [{}]".format(value))
                echo("Unexpected traceback: [{}]".format(traceback))
                self.update_status(Migration.STATE_FAILED)
                raise
            else:
                echo("Migration {} Successful".format(self.migration_name))
                self.update_status(Migration.STATE_COMPLETED)
        elif self.status == Migration.STATE_PROCESSING:
            echo("{} is currently being processed".format(self.migration_name))
        elif self.status == Migration.STATE_COMPLETED:
            echo("{} has already been processed".format(self.migration_name))
        elif self.status == Migration.STATE_FAILED:
            echo("{} has already been processed, and failed - best to restart".format(self.migration_name))


    def run(self):
        """Should be implemented by subclass"""
        raise NotImplementedError("This is an abstract class")


class QuerySet(object):

    def __init__(self, database, mongodump_options):
        self.database = database
        self.mongodump_options = mongodump_options
        self.touched_collections = []

    @property
    def application_collection_names(self):
        """returns an array of  names of all non system tables in the database"""
        system_table_re = re.compile("system\.")

        return [col_name for col_name in self.database.collection_names() if not system_table_re.match(col_name)]

    def dump_collection(self, collection_name, query=None):
        self.touched_collections.append(collection_name)
        execution_array = ['mongodump']

        collection_options = copy(self.mongodump_options)
        collection_options['-c'] = collection_name

        if query:
            collection_options['-q'] = str(query)

        for option in collection_options:
            execution_array.extend([option, collection_options[option]])
        echo("Executing: {}".format(execution_array))
        subprocess.call(execution_array)

    def run(self):
        """Should be implemented by the subclass"""
        raise NotImplementedError("This is an abstract class")

    def only(self):
        """if you want to limit the collections override this and return an array of collection names"""
        return None

    def exclude(self):
        """if you want to limit the collections override this and return an array of collection names"""
        return None

    @property
    def additional_collections(self):
        """returns the collections not specified in the query_set factoring `only` and `exclude`"""
        include_set = self.only() or self.application_collection_names
        include_set = set(include_set) - set(self.touched_collections)

        if self.exclude():
            include_set = include_set - set(self.exclude())

        return include_set

    def execute(self):
        self.run()

        for collection_name in self.additional_collections:
            self.dump_collection(collection_name)


class MigrationHistoryStorage(object):
    """Contract for persistence implementations to adhere to"""
    @classmethod
    def find_or_create_by_key(cls, migration_key):
        """Should return a MigrationHistory object"""
        raise NotImplementedError('Abstract Class')


class MigrationHistory(object):
    """Contract for persistence implementations to adhere to"""
    def __init__(self, **kwargs):
        self.key = kwargs.get('key')
        self.state = kwargs.get('state')
        self.processed_at = kwargs.get('processed_at')



import os
from datetime import datetime

# 3rd Party Imports
import boto
from boto.s3.key import Key
from click import echo

from .utils import temp_directory, zipdir
from .local import local_restore
from .mongo import dump_db


def get_s3_bucket(s3_settings):
    conn = boto.connect_s3(s3_settings['aws_access_key_id'], s3_settings['aws_secret_access_key'])
    bucket = conn.get_bucket(s3_settings['bucket_name'])
    return bucket


def generate_uniqueish_key(s3_settings, environment, name_prefix):
    bucket = get_s3_bucket(s3_settings)

    if name_prefix and name_prefix != '':
        name_base = name_prefix
    else:
        name_base = environment['db_name']

    name_attempt = "{}__{}.dmp.zip".format(name_base, datetime.utcnow().strftime("%Y_%m_%d"))

    key = bucket.get_key(name_attempt)

    if not key:
        key = Key(bucket)
        key.key = name_attempt
        return key
    else:
        counter = 1
        while True:
            counter += 1
            name_attempt = "{}__{}_{}.dmp.zip".format(name_base,
                                          datetime.utcnow().strftime("%Y_%m_%d"), counter)

            if bucket.get_key(name_attempt):
                continue
            else:
                key = Key(bucket)
                key.key = name_attempt
                return key


def backup_to_s3(environment, s3_settings, name, query_set_class):

    dump_path = dump_db(environment, QuerySet=query_set_class)
    zipf = zipdir(dump_path)

    key = generate_uniqueish_key(s3_settings, environment, name)

    bytes_written = key.set_contents_from_filename(zipf.filename)

    # 4) print out the name of the bucket
    echo("Wrote {} bytes to s3".format(bytes_written))


def s3_restore(key, to_enviornment):

    with temp_directory() as temp_dir:
        zip_path = os.path.join(temp_dir, 'MongoDump.zip')
        key.get_contents_to_filename(zip_path)
        local_restore(zip_path, to_enviornment)


def s3_backups(s3_config):
    """ a dict of key.name: key
    """
    bucket = get_s3_bucket(s3_config)

    buckets = {}
    for key in bucket.get_all_keys():
        buckets[key.name] = key

    return buckets



import os
import zipfile
import re
import shutil
from tempfile import mkdtemp
from contextlib import contextmanager

from click import echo

CAMEL_PAT = re.compile(r'([A-Z])')
UNDER_PAT = re.compile(r'_([a-z])')


@contextmanager
def temp_directory():
    temp_dir = mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


def camel_to_underscore(name):
    return CAMEL_PAT.sub(lambda x: '_' + x.group(1).lower(), name)


def underscore_to_camel(name):
    return UNDER_PAT.sub(lambda x: x.group(1).upper(), name.capitalize())


def sizeof_fmt(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0


def zipdir(dump_path):
    def _zipdir(path, zip):
        for root, dirs, files in os.walk(path):
            for file in files:
                zip.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), root))

    zipf = zipfile.ZipFile('MongoDump.zip', 'w')
    _zipdir(dump_path, zipf)
    zipf.close()
    return zipf


def exit_with_message(message):
    echo()
    echo(message)
    echo()
    exit()



# Core Imports
import os
import re
import sys
from importlib import import_module
from six import iteritems

# 3rd Party Imports
import click

from click import echo, progressbar

#exposing echo and progressbar as utilities for migrations
echo = echo
progressbar = progressbar

# Local Imports
from .models import Migration, QuerySet
from .local import local_restore, local_backups, backup_localy
from .s3 import get_s3_bucket, generate_uniqueish_key, backup_to_s3, s3_restore, s3_backups
from .migrations import generate_migration_name, create_package_if_necessary, find_migrations
from .query_sets import querysets, generate_queryset_name

from .mongo import MongoMigrationHistory, MongoBackedMigration, \
    establish_datastore_connection, \
    restore as restore_mongo_db, \
    copy_db as copy_mongo_db, \
    drop as drop_mongo_db

from .utils import temp_directory, camel_to_underscore, \
    underscore_to_camel, sizeof_fmt, exit_with_message

from .templates import MIGRATION_TEMPLATE, CONFIG_TEMPLATE, QUERYSET_TEMPLATE


class Config(object):

    def __init__(self):
        self.migration_directory = './migrations'
        self.queryset_directory = './querysets'
        self.config_directory = None

    def configure_from_settings_file(self):
        try:
            sys.path.append(os.getcwd())

            settings = import_module('migrations.settings')
        except ImportError:
            exit_with_message("Could not find your settings.py file -- did you run monarch init?")

        if not hasattr(settings, 'ENVIRONMENTS'):
            exit_with_message('Configuration file should have a ENVIRONMENTS method set')
        else:
            self.environments = settings.ENVIRONMENTS

        if hasattr(settings, 'BACKUPS'):
            self.backups = settings.BACKUPS

            if 'S3' in self.backups and 'LOCAL' in self.backups:
                exit_with_message('BACKUPS Setting has both LOCAL and S3 config -- choose one')
            elif 'S3' not in self.backups and 'LOCAL' not in self.backups:
                exit_with_message('BACKUPS are configured, but S3 or LOCAL is not defined -- please define one')
            elif 'S3' in self.backups:
                s3_config = self.backups['S3']

                required_config = ['bucket_name', 'aws_access_key_id', 'aws_secret_access_key']

                missing_config = []

                for item in required_config:
                    if item not in s3_config:
                        missing_config.append(item)

                if missing_config:
                    msg = "Missing [] items in S3 section of your settings.py".format(", ".join(missing_config))
                    exit_with_message(msg)

pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
@pass_config
@click.pass_context
def cli(ctx, config):
    """ Your friendly migration manager

        To get help on a specific function you may append --help to the function
        i.e.
        monarch generate --help
    """
    if ctx.invoked_subcommand != 'init':
        config.configure_from_settings_file()


@cli.command()
@click.argument('name')
@pass_config
def generate(config, name):
    """
    Generates a migration file.  pass it a name.  execute like so:

    monarch generate [migration_name]

    i.e.

    monarch generate add_indexes_to_user_collection

    """
    create_package_if_necessary(config.migration_directory)
    migration_name = generate_migration_name(config.migration_directory, name)
    class_name = "{}Migration".format(underscore_to_camel(name))
    output = MIGRATION_TEMPLATE.format(migration_class_name=class_name, base_class='MongoBackedMigration')
    with open(migration_name, 'w') as f:
        f.write(output)
    click.echo("Generated Migration Template: [{}]".format(migration_name))

@cli.command()
@click.argument('name')
@pass_config
def generate_query_set(config, name):
    """
    Generates a query_set file.  pass it a name.  execute like so:

    monarch generate_query_set [queryset_name]

    i.e.

    monarch generate_query_set data_for_account

    """
    create_package_if_necessary(config.queryset_directory)
    queryset_name = generate_queryset_name(config.queryset_directory, name)
    class_name = "{}QuerySet".format(underscore_to_camel(name))
    output = QUERYSET_TEMPLATE.format(queryset_class_name=class_name, base_class='QuerySet')
    with open(queryset_name, 'w') as f:
        f.write(output)
    click.echo("Generated Query Set: [{}]".format(queryset_name))



@cli.command(name='list_migrations')
@click.argument('environment')
@pass_config
def lizt(config, environment):
    """ Lists the migrations and the status against the specified environemnt

    """
    if environment not in config.environments:
        exit_with_message("Environment not described in settings.py")

    migrations_on_file_system = find_migrations(config)

    establish_datastore_connection(config.environments[environment])

    if migrations_on_file_system:
        click.echo("Here are the migrations:")
        echo("{:50} {}".format('MIGRATIONS', 'STATUS'))
        for migration_name in migrations_on_file_system:
            migration_meta = MongoMigrationHistory.find_by_key(migration_name)
            if migration_meta:
                echo("{:50} {}".format(migration_name, migration_meta.state))
            else:
                echo("{:50} NOT RUN".format(migration_name))

    else:
        click.echo("No pending migrations")


@cli.command()
@click.argument('environment')
@pass_config
def migrate(config, environment):
    """
    Runs all migrations that have yet to have run.
    :return:
    """
    if environment not in config.environments:
        exit_with_message("Environment not described in settings.py")

    check_for_hazardous_operations(config, environment)

    # 1) Find all migrations in the migrations/ directory
    # key = name, value = MigrationClass
    migrations = find_migrations(config)
    if migrations:
        establish_datastore_connection(config.environments[environment])
        for k, migration_class in iteritems(migrations):
            migration_instance = migration_class()

            # 3) Run the migration -- it will only run if it has not yet been run yet
            migration_instance.process()
    else:
        click.echo("No migrations exist")



@cli.command()
@click.argument('migration_name')
@click.argument('environment')
@pass_config
def migrate_one(config, migration_name, environment):
    """
    Runs all migrations that have yet to have run.
    :return:
    """
    if environment not in config.environments:
        exit_with_message("Environment not described in settings.py")

    check_for_hazardous_operations(config, environment)

    establish_datastore_connection(config.environments[environment])

    # 1) Find all migrations in the migrations/ directory
    # key = name, value = MigrationClass
    migration = find_migration(config, migration_name)
    migration.process(force=True)


def find_migration(config, migration_name):

    migrations = find_migrations(config)
    migration_class = migrations[migration_name]
    return migration_class()



@cli.command()
@click.option('--migration-directory', default='./migrations', help='path to where you want to store your migrations')
def init(migration_directory):
    """ Generates a default setting file.

        It will it in ./migrations and will create the package if it does not exist

    """

    create_package_if_necessary(migration_directory)
    settings_file = os.path.join(os.path.abspath(migration_directory), 'settings.py')

    if os.path.exists(settings_file):
        click.confirm("A settings file already exists.  Are you sure you want to overwrite it?", abort=True)

    with open(settings_file, 'w') as f:
        f.write(CONFIG_TEMPLATE)

    msg = """We just created a shinny new configuration file for you.  You can find it here:

    {}

    You are encouraged to open it up and modify it for your needs
    """.format(settings_file)

    echo(msg)


@cli.command()
@click.option('--query-set', help='provide optional query-set filter, default is the entire db')
@click.argument('from_to')
@pass_config
def copy_db(config, from_to, query_set):
    """ Copys a database and imports into another database

        Example

        monarch import_db production:local
        monarch import_db staging:local

    """
    if ':' not in from_to:
        exit_with_message("Expecting from:to syntax like production:local")

    from_db, to_db = from_to.split(':')

    check_for_hazardous_operations(config, to_db)

    if config.environments is None:
        exit_with_message('Configuration file should have a ENVIRONMENTS set')

    if from_db not in config.environments:
        exit_with_message('Environments does not have a specification for {}'.format(from_db))

    if to_db not in config.environments:
        exit_with_message('Environments does not have a specification for {}'.format(to_db))

    query_set_class = None
    if query_set:
        if query_set not in querysets(config):
            exit_with_message('Could not find specified query_set in your queryset folder')
        else:
            query_set_class = querysets(config)[query_set]

    if click.confirm('Are you SURE you want to copy data from {} into {}?'.format(from_db, to_db)):
        echo()
        echo("Okay, you asked for it ...")
        echo()
        copy_mongo_db(config.environments[from_db],
                      config.environments[to_db],
                      query_set_class)


@cli.command()
@click.argument('environment')
@pass_config
def drop_db(config, environment):
    """ drops the database -- ARE YOU SURE YOU WANT TO DO THIS
    """

    check_for_hazardous_operations(config, environment)

    drop_mongo_db(config.environments[environment])


@cli.command()
@click.argument('environment')
@click.option('--name', help='name to prefix the backup with')
@click.option('--query-set', help='provide optional query-set filter, default is the entire db')
@pass_config
def backup(config, environment, name, query_set):
    """ Backs up a given datastore
        It is configured in the BACKUPS section of settings
        You can back up locally or to S3

        use --name if you want to specify a name, otherwise it will use your environment name

    """
    env_name = environment

    if not hasattr(config, 'backups'):
        exit_with_message('BACKUPS not configured, exiting')

    environment = confirm_environment(config, env_name)

    query_set_class = None
    if query_set:
        if query_set not in querysets(config):
            exit_with_message('Could not find specified query_set in your queryset folder')
        else:
            query_set_class = querysets(config)[query_set]

    if 'LOCAL' in config.backups:
        backup_localy(environment, config.backups['LOCAL'], name, query_set_class)
    elif 'S3' in config.backups:
        backup_to_s3(environment, config.backups['S3'], name, query_set_class)
    else:
        exit_with_message('BACKUPS not configured, exiting')


@cli.command()
@pass_config
def list_backups(config):
    """ Lists available backups
    """
    if config.backups is None:
        exit_with_message('BACKUPS not configured, exiting')

    if 'LOCAL' in config.backups:
        list_local_backups(config.backups['LOCAL'])
    elif 'S3' in config.backups:
        list_s3_backups(config.backups['S3'])
    else:
        exit_with_message('BACKUPS not configured, exiting')


@cli.command()
@pass_config
def list_environments(config):
    """ Lists Configured Environments
    """
    if config.environments:
        for env in config.environments:
            echo("{:40}: {}".format(env, config.environments[env]))
    else:
        echo()
        echo("Yikes you have no environments set up -- you should set some up.  Maybe rerun monarch init")
        echo()


@cli.command()
@pass_config
def list_query_sets(config):
    """ Lists Query Sets
    """
    query_sets_on_file_system = querysets(config)

    if query_sets_on_file_system:
        click.echo("Here are the query sets:")
        echo('QUERY SETS')
        for queryset_name in query_sets_on_file_system:
            echo(queryset_name)

    else:
        click.echo("No query sets created")


@cli.command()
@click.argument('from_to')
@pass_config
def restore(config, from_to):
    """ Restores a backup into a destination database.  Provide a dump name that you can get from

        monarch list_backups

        Example

        monarch restore adid-development__2014_06_18.dmp.zip:development

    """
    if ':' not in from_to:
        exit_with_message("Expecting from:to syntax like production:local")

    backup, to_db = from_to.split(':')

    check_for_hazardous_operations(config, to_db)

    if config.environments is None:
        exit_with_message('Configuration file should have a ENVIRONMENTS set')

    if to_db not in config.environments:
        exit_with_message('Environments does not have a specification for {}'.format(to_db))

    if backup not in backups(config):
        exit_with_message('Can not find backup {}, run monarch list_backups to see your options'.format(backup))

    msg = 'Are you SURE you want to restore backup into into {}? It will delete the database first'.format(to_db)
    if click.confirm(msg):
        echo()
        echo("Okay, you asked for it ...")
        echo()
        restore_db(config, backups(config)[backup], config.environments[to_db])


def confirm_environment(config, env_name):
    if env_name not in config.environments:
        exit_with_message("{} is not in settings.  Exiting ...".format(env_name))
    else:
        return config.environments[env_name]


def list_local_backups(local_config):

    _local_backups = local_backups(local_config)
    if _local_backups:
        for backup in _local_backups:
            echo("{:50} {}".format(backup, sizeof_fmt(os.path.getsize(_local_backups[backup]))))
    else:
        echo()
        echo('You have not backups yet -- make some?  monarch backup <env_name>')
        echo()


def list_s3_backups(s3_settings):
    _s3_backups = s3_backups(s3_settings)
    if _s3_backups:
        for backup in _s3_backups:
            echo("{:50} {}".format(backup, sizeof_fmt(_s3_backups[backup].size)))
    else:
        echo()
        echo('You have not backups yet -- make some?  monarch backup <env_name>')
        echo()


def restore_db(config, path_or_key, to_environment):
    """unzips the file then runs a restore"""

    if config.backups is None:
        exit_with_message('BACKUPS not configured, exiting')

    if 'LOCAL' in config.backups:
        return local_restore(path_or_key, to_environment)
    elif 'S3' in config.backups:
        return s3_restore(path_or_key, to_environment)
    else:
        exit_with_message('BACKUPS not configured, exiting')

    echo()
    echo("Rock and roll that seemed to go well -- Nice work")
    echo()


def backups(config):
    """returns a dictionary of {backup_name: backup_path}"""
    if config.backups is None:
        exit_with_message('BACKUPS not configured, exiting')

    if 'LOCAL' in config.backups:
        return local_backups(config.backups['LOCAL'])
    elif 'S3' in config.backups:
        return s3_backups(config.backups['S3'])
    else:
        exit_with_message('BACKUPS not configured, exiting')


def test_for_human():
    from random import randint
    int_1 = randint(0, 100)
    int_2 = randint(0, 100)

    sum = int_1 + int_2

    value = click.prompt('What is the sum of {} and {}?'.format(int_1, int_2), type=int)

    if value != sum:
        exit_with_message('You have chosen poorly')
    else:
        echo("Good enough for me -- continue")

ensure_smarter_than_a_4_year_old = test_for_human


def check_for_hazardous_operations(config, env_name):

    if env_name not in config.environments:
        exit_with_message("Environment not described in settings.py")

    env = config.environments[env_name]
    db_name = env['db_name']
    db_host = env['host']

    ends_with_dot_local = re.compile("local$")

    def looks_like_a_remote_host(host_name):
        if db_host in ('localhost', '127.0.0.1'):
            return False

        if 'localhost' in db_host:
            return False

        if '127.0.0.1' in db_host:
            return False

        if ends_with_dot_local.search(db_host) is not None:
            return False

        #not sure so assuming remote
        return True

    dangerous = (env_name == 'production' or looks_like_a_remote_host(db_host))

    if dangerous:
        echo('You are about to perform a potentially hazardous operation. \n\nenv: [{}] db_host [{}]\nComputer Breath-o-lizer test:'.format(env_name, db_host))
        ensure_smarter_than_a_4_year_old()



MIGRATION_TEMPLATE = '''
from monarch import echo
from monarch import {base_class}

# Optionally
# from monarch import progressbar
#
# Usage:
# with progressbar(all_the_users_to_process,
#                  label='Modifying user accounts',
#                  length=number_of_users) as bar:
#    for user in bar:
#        modify_the_user(user)
#
#
# Note: Using echos (or prints) will mess up the way this looks
#
# More info: http://click.pocoo.org/utils/#showing-progress-bars


class {migration_class_name}({base_class}):

    def run(self):
        """Write the code here that will migrate the database from one state to the next
            No Need to handle exceptions -- we will take care of that for you
        """
        raise NotImplementedError


'''

QUERYSET_TEMPLATE = '''
from click import prompt
from monarch import {base_class}


class {queryset_class_name}({base_class}):

    def run(self):
        """Write the code here that will perform the mongo dump of the collections that you care about
        """
        raise NotImplementedError


'''

CONFIG_TEMPLATE = """
# monarch settings file, generated by monarch init
# feal free to edit it with your application specific settings

ENVIRONMENTS = {
    'production': {
        'host': 'your-host:12345,your-other-host:12347',
        'db_name': 'your-db-name',
        'username': 'asdf',
        'password': 'asdfdf',
        'sslCAFile': '/path/to/production.pem'
    },
    'development': {
        'host': 'your-host:12345',
        'db_name': 'your-db-name',
        'username': 'asdf',
        'password': 'asdfdf'
    },
}


# If you want to use the backups feature uncomment and fill out the following:
# BACKUPS = {
#     'S3': {
#         'bucket_name': 'your_bucket_name',
#         'aws_access_key_id': 'aws_access_key_id',
#         'aws_secret_access_key': 'aws_secret_access_key',
#     }
# }

# OR

# BACKUPS = {
#     'LOCAL': {
#         'backup_dir': 'path_to_backups',
#     }
# }


"""


import os
from setuptools import setup


def get_packages():
    # setuptools can't do the job :(
    packages = []
    for root, dirnames, filenames in os.walk('monarch'):
        if '__init__.py' in filenames:
            packages.append(".".join(os.path.split(root)).strip("."))

    return packages

setup(
    name='monarch',
    version='0.1.11',
    description='The un-migration migration tool',
    url='http://github.com/jtushman/monarch',
    author='Jonathan Tushman',
    author_email='jonathan@zefr.com',
    packages=get_packages(),
    install_requires=[
        'Click>2.0',
        'jinja2',
        'mongoengine',
        'boto',
    ],
    tests_require=['nose'],
    test_suite='nose.collector',
    entry_points='''
        [console_scripts]
        monarch=monarch:cli
    ''',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
)
